#!/bin/bash
# This script will work with one squid or up to 4 squids on the same http port.
# The number of squids is determined by the existence of cache directories
# as follows.  The main path to the cache directories is determined by the
# cache_dir option in squid.conf.  To run multiple squids, create directories
# of the form
#   $cache_dir/squid$N
# where N goes from 0 to the number of squids minus 1.  cache_log directories
# of the same form will be used, and created if needed.  Note that the
# cache_log option in squid.conf is a file, not a directory, so the squid$N
# is up one level:
#   cache_log_file=`basename $cache_log`
#   cache_log_dir=`dirname $cache_log`/squid$N
#   cache_log=$cache_log_dir/$cache_log_file
# The access_log should be in the same directory as the cache_log (unless it
# is set to "none"), and the pid_filename also will be in similarly named
# directories (the same directories as the cache_log is a good choice but not
# required, and the directories will be created if needed).
# Each squid also listens separately on port 3127-$N for http requests
#  in addition to the shared port 3128, and each squid listens on
#  port 3401+$N for snmp queries.  Only the first squid listens on 3130
#  for icp.  (All of those starting numbers can be overridden in squid.conf,
#  they're just the defaults.)  Each squid after the first one uses the
#  first one as a parent, so they don't all contact the upstream servers
#  separately.

. /etc/init.d/functions

RETVAL=0

CONF_DIR=/etc/squid
SBIN_DIR=/usr/sbin
AWSTATS_SCRIPT=/etc/awstats/run_awstats.sh
#size in bytes at which rotateiflarge will rotate access.log
LARGE_ACCESS_LOG=${LARGE_ACCESS_LOG:-1000000000}

CONF_FILE=$CONF_DIR/squid.conf

if [ `id -u` = 0 ]; then
  echo "ERROR: `basename $0` not allowed to be run as root" >&2
  exit 1
fi

if [ ! -d "$CONF_DIR" ] || [ ! -w "$CONF_DIR" ]; then
 echo "ERROR: $CONF_DIR does not exist or is not writable" >&2
 exit 1
fi

# see if squid.conf needs to be regenerated
if [ $CONF_FILE.frontierdefault -nt $CONF_FILE ] ||
	[ $CONF_DIR/customize.sh -nt $CONF_FILE ]; then
 if [ -f "$CONF_FILE" ] && [ -n "`find "$CONF_FILE" -perm -0200 2>/dev/null`" ]; then
  echo "ERROR: $CONF_FILE is writable and was about to be overwritten" >&2
  echo "  This may indicate that squid was upgraded from an older version" >&2
  echo "  that was not based on customize.sh." >&2
  echo "  Remove or rename it, or chmod it to be non-writable" >&2
  exit 1
 fi
 echo "Generating $CONF_FILE"
 rm -f $CONF_FILE.new
 if $CONF_DIR/customize.sh <$CONF_FILE.frontierdefault >$CONF_FILE.new; then
  rm -f $CONF_FILE.old
  if [ -f $CONF_FILE ]; then
   echo "Saving old configuration as $CONF_FILE.old"
   mv $CONF_FILE $CONF_FILE.old
  fi
  mv $CONF_FILE.new $CONF_FILE
  chmod 444 $CONF_FILE
 else
  exit $?
 fi
fi

# assume there are no spaces in these file paths
CACHE_DIR=`awk '$1 == "cache_dir" {x=$3} END{print x}' $CONF_FILE`
CACHE_LOG=`awk '$1 == "cache_log" {x=$2} END{print x}' $CONF_FILE`
ACCESS_LOG=`awk '$1 == "access_log" {x=$2} END{print x}' $CONF_FILE`
PID_FILENAME=`awk '$1 == "pid_filename" {x=$2} END{print x}' $CONF_FILE`

if [ -z "$CACHE_DIR" ]; then
 echo "ERROR: cache_dir not found in $CONF_FILE" >&2
 exit 1
fi
if [ "$CACHE_DIR" == "/" ]; then
 echo "ERROR: cache_dir not allowed to be '/' in $CONF_FILE" >&2
 exit 1
fi
if [ ! -d "$CACHE_DIR" ] || [ ! -w "$CACHE_DIR" ]; then
 echo "ERROR: cache_dir $CACHE_DIR does not exist or is not writable" >&2
 exit 1
fi
if [ -z "$CACHE_LOG" ]; then
 echo "ERROR: cache_log not found in $CONF_FILE" >&2
 exit 1
fi
CACHE_LOG_DIR=`dirname "$CACHE_LOG"`
if [ ! -d "$CACHE_LOG_DIR" ] || [ ! -w "$CACHE_LOG_DIR" ]; then
 echo "ERROR: cache_log directory $CACHE_LOG_DIR does not exist or is not writable" >&2
 exit 1
fi
if [ -z "$ACCESS_LOG" ]; then
 echo "ERROR: access_log not found in $CONF_FILE" >&2
 exit 1
fi
if [ "$ACCESS_LOG" != "none" ]; then
 ACCESS_LOG_DIR=`dirname "$ACCESS_LOG"`
 if [ "$ACCESS_LOG_DIR" != "$CACHE_LOG_DIR" ]; then
  echo "ERROR: access_log directory $ACCESS_LOG_DIR is not the same as the cache_log directory $CACHE_LOG_DIR" >&2
  exit 1
 fi
fi
if [ -z "$PID_FILENAME" ]; then
 echo "ERROR: pid_filename not found in $CONF_FILE" >&2
 exit 1
fi
PID_FILENAME_DIR=`dirname "$PID_FILENAME"`
if [ ! -d "$PID_FILENAME_DIR" ] || [ ! -w "$PID_FILENAME_DIR" ]; then
 echo "ERROR: pid_filename directory $PID_FILENAME_DIR does not exist or is not writable" >&2
 exit 1
fi

LOGFILE_ROTATE=`awk '$1 == "logfile_rotate" {x=$2} END{print x}' $CONF_FILE`
if [ -z "$LOGFILE_ROTATE" ]; then
 LOGFILE_ROTATE=10
fi

squid_dirs()
{
 # if $NUMSQUIDS is 1, echo the parameter, otherwise echo the parameter N
 #   times with /squid$N after the parameter, where N is from 0 to $NUMSQUIDS-1

 if [ $NUMSQUIDS = 1 ]; then
  echo $1
  return
 fi
 typeset N=0
 typeset SQUID_DIRS=""
 while [ $N -lt $NUMSQUIDS ]; do
  SQUID_DIRS="$SQUID_DIRS $1/squid$N"
  let N=$N+1
 done
 echo $SQUID_DIRS
}

#see how many squid cache directories exist (up to 4)
NUMSQUIDS=0
for D in `NUMSQUIDS=4 squid_dirs $CACHE_DIR`; do
 if [ ! -d $D ]; then
  break
 fi
 let NUMSQUIDS=$NUMSQUIDS+1
done

SQUIDEXE=$SBIN_DIR/squid
if [ $NUMSQUIDS = 0 ]; then
 NUMSQUIDS=1
 SQUID=$SQUIDEXE
 PLURAL=""
else
 PLURAL="s"
 if [ -z "$SETSQUIDAFFINITY" ]; then
  # When there's more than one squid (and no more cores than squids), default
  #  to set the affinity of each squid to one core for best performance.
  # Measurements showed a ~15% throughput boost per core when cpu-limited.
  let NUMCORES="$(grep ^processor /proc/cpuinfo|tail -1|sed 's/.*: //') + 1"
  if [ "$NUMCORES" -ge "$NUMSQUIDS" ]; then
   SETSQUIDAFFINITY=true
  fi
 fi

 #create corresponding $CACHE_LOG_DIR and $PID_FILENAME_DIR directories
 #  if needed
 for D in $(squid_dirs $CACHE_LOG_DIR) $(squid_dirs $PID_FILENAME_DIR); do
  if [ ! -d $D ]; then
   mkdir $D
  fi
 done

 # create the squid.conf files for each squid
 HTTP_PORT=`awk '$1 == "http_port" {print $2;exit}' $CONF_FILE`
 SNMP_PORT=`awk '$1 == "snmp_port" {x=$2} END{print x}' $CONF_FILE`
 ICP_PORT=`awk '$1 == "icp_port" {x=$2} END{print x}' $CONF_FILE`
 #visible_hostname is the name that shows up in X-Cache and Via headers
 VISIBLE_HOSTNAME=`awk '$1 == "visible_hostname" {x=$2} END{print x}' $CONF_FILE`
 HOSTNAME=`hostname`
 SQUID="$(dirname $(dirname $0))/sbin/multisquid $SQUIDEXE $CONF_DIR $NUMSQUIDS ${HTTP_PORT:-3128}"
 N=0
 while [ $N -lt $NUMSQUIDS ]; do
  SEDCMDS="-e \"s,$CACHE_DIR,$CACHE_DIR/squid$N,\""
  if [ "$CACHE_LOG_DIR" != "$CACHE_DIR" ]; then
   SEDCMDS="$SEDCMDS -e \"s,$CACHE_LOG_DIR/,$CACHE_LOG_DIR/squid$N/,\""
  fi
  if [ "$PID_FILENAME_DIR" != "$CACHE_DIR" ] && [ "$PID_FILENAME_DIR" != "$CACHE_LOG_DIR" ]; then
   SEDCMDS="$SEDCMDS -e \"s,$PID_FILENAME_DIR/,$PID_FILENAME_DIR/squid$N/,\""
  fi
  NEWCONF=$CONF_DIR/.squid-$N.conf
  rm -f $NEWCONF
  eval sed $SEDCMDS $CONF_DIR/squid.conf | awk '
   BEGIN{print "# DO NOT EDIT -- Automatically generated by '$0'";print}
   /^[ \t#]*http_port[ \t]/ {
    if (!got_http_port) {
     #add additional http_port after first one specified, or after first
     #   commented http_port if none specified
     portnum=0
     if (substr($0,1,1) != "#") portnum=$2
     else if ("'$HTTP_PORT'" == "") {portnum=3128; print}
     if (portnum != 0) {
      print "#first port overridden by command line -I"
      print "http_port", portnum
      # have each squid also listen on their own port, the first one to
      # be parent of others and the others for forced cache reloads
      print "http_port", portnum - ('$N' + 1)
      got_http_port=1
      next
     }
    }
    else if ((substr($0,1,1) != "#") && ("'$N'" != 0)) {
     # only first squid listens on other specified http ports, skip this one
     next
    }
   }
   /^[ \t#]*snmp_port[ \t]/ {
    #replace snmp_port with separate number for each squid
    portnum=0
    if (substr($0,1,1) != "#") portnum=$2
    else if ("'$SNMP_PORT'" == "") {portnum=3401; print}
    if (portnum != 0) {
     portnum=portnum + '$N'
     print "snmp_port", portnum
     next
    }
   }
   /^[ \t#]*icp_port[ \t]/ {
    #only first squid listens on icp port
    portnum=0
    if (substr($0,1,1) != "#") portnum=$2
    else if ("'$ICP_PORT'" == "") {portnum=3130; print}
    if (portnum != 0) {
     if ("'$N'" != 0) portnum=0  # disables the icp port
     print "icp_port", portnum
     next
    }
   }
   /^[ \t#]*cache_peer[ \t]/ {
    #any squid after first one go only to first one as a cache_peer parent
    if ("'$N'" != 0) {
     if (substr($0,1,1) == "#") {
      if (!got_cache_peer) {
       #insert the line after the first commented-out one
       print
       portnum='${HTTP_PORT:-3128}' - 1
       print "cache_peer localhost parent", portnum, "0 no-query"
       got_cache_peer=1
       next
      }
     }
     else {
      #skip any other uncommented cache_peer options on non-first squids
      next
     }
    }
   }
   /^ *update_headers[ \t]/ {
    if (("'$N'" == 0) && ("'$NUMSQUIDS'" > 0)) {
     # this has to be on for squids that feed other squids
     print "update_headers on"
    }
    else {
     # take default
     print
    }
    next
   }
   $1 == "visible_hostname" || /TAG: visible_hostname/ {
    hostname=""
    if (substr($0,1,1) != "#") hostname=$2
    else if ("'$VISIBLE_HOSTNAME'" == "") {hostname="'$HOSTNAME'"; print}
    if (hostname != "") {
     print "visible_hostname",hostname "/'$N'"
    }
   }
   {print}
   ' >$NEWCONF
  chmod a-w $NEWCONF
  let N=$N+1
 done
fi

last_logline_has()
{
  # if first parameter is 1, see if any of the last 5 lines of any of the
  #  cache logs of squids contains the other parameters; if the first parameter
  #  is 0, see if true for all of the cache logs
  typeset ANY LDIR LFILE
  ANY=$1
  shift
  for LDIR in $(squid_dirs $CACHE_LOG_DIR); do
    LFILE=$LDIR/`basename $CACHE_LOG`
    if tail -5 $LFILE | egrep -q "$*"; then
      if [ "$ANY" = 1 ]; then
	# found one match, return true
        return 0
      fi
    else
      if [ "$ANY" = 0 ]; then
	# found one non-match, return false
        return 1
      fi
    fi
  done
  # return true for ANY=0 (all matched), false for ANY=1 (none matched)
  return $ANY
}

isrunning()
{
 # return 0 if all running, else 1
 for D in `squid_dirs $PID_FILENAME_DIR`; do
  typeset MY_PID_FILENAME="$D/$(basename $PID_FILENAME)"
  if [ ! -f "$MY_PID_FILENAME" ]; then
   return 1
  fi
  typeset PID="`cat <$MY_PID_FILENAME 2>/dev/null`"
  if [ ! -n "$PID" ]; then
   rm -f "$MY_PID_FILENAME"
   return 1
  fi
  # Need to make sure that $PID actually refers to squid and not some
  #   other program after a system failure & restart.
  # Can't use pidof, because nightly prelink confuses it when $SBIN_DIR
  #   is a system directory.
  # Can't use ls -iL on /proc/$PID/exe, because that shows the inode
  #   of the executable before prelink messed with it, so instead extract
  #   the name of the symlink target and do -iL on that.
  typeset PIDEXE="`ls -l /proc/$PID/exe 2>/dev/null|sed 's,.*-> ,,'`"
  if [ -n "$PIDEXE" ]; then
   # Compare the inodes of $PID's executable and $SQUID to see if they're
   #  the same file.
   # Note that the syntax ${VAR%% *} removes everything in $VAR after a blank.
   # RHEL6-based systems can have a " (deleted)" as part of $PIDEXE.
   typeset PIDINODE="`ls -iL ${PIDEXE%% *} 2>/dev/null`"
   typeset SQUIDINODE="`ls -iL ${SQUIDEXE%% *}`"
   if [ "${PIDINODE%% *}" = "${SQUIDINODE%% *}" ]; then
    continue
   fi
  fi
  return 1
 done
 return 0
}

cleancache()
{
 for D in `squid_dirs $CACHE_DIR`; do
  if [ -f $D/swap.state ]; then
   # a cache exists
   echo "Checking if $NUMSQUIDS Frontier Squid${PLURAL} running..."
   if $SQUID -k parse; then
    if isrunning; then
     echo "Squid${PLURAL} running, doing nothing"
     exit 1
    fi
   else
    RETVAL=$?
    echo "ERROR parsing $CONF_FILE"
    exit $RETVAL
   fi
   echo "Start clearing Cache${PLURAL} of $NUMSQUIDS Frontier Squid${PLURAL} in the background... "
   $SQUID -k shutdown 2>/dev/null||true  # in case all are not shut down
   for D in `squid_dirs $CACHE_DIR`; do
    rm -rf $D/.deleting
    mkdir -p $D/.deleting
    mv $D/* $D/.deleting 2>/dev/null
   done
   nohup $SHELL -c "for D in `squid_dirs $CACHE_DIR`; do
    rm -rf \$D/.deleting
   done" </dev/null >/dev/null 2>&1 &
   echo "started"
   break
  fi
 done
 echo "Initializing Cache${PLURAL}..."
 $SQUID -z
 sleep 3
}

start_squid()
{
 echo "Starting $NUMSQUIDS Frontier Squid${PLURAL}... "
 $SQUID -DFS
 # it can take a while to check out a large cache


 LIMIT=60   # Number of Tries
 a=1
 while [ $a -le "$LIMIT" ]
 do
  a=$(($a+1)) 
  sleep 10
  isrunning
  RETVAL=$?
  if [ $RETVAL != 0 ] ; then
   echo "Frontier Squid start failed!!!"
   start_squid_fail
   RETVAL=$?
   break
  else
   if last_logline_has 0 "storeLateRelease: released" ;then
    break
    RETVAL=0
   fi
  fi
 done

 if [ "$SETSQUIDAFFINITY" = true ]; then
  let AFFINITYCORE=0
  for D in `squid_dirs $PID_FILENAME_DIR`; do
   taskset -cp $AFFINITYCORE `cat $D/$(basename $PID_FILENAME)` >/dev/null
   let AFFINITYCORE+=1
  done
 fi

 sleep 2
 echo "done"

 [ $RETVAL == 0 ]
}

start_squid_fail()
{
 RETVAL=0
 $SQUID -k shutdown 2>/dev/null|| true  # in case all are not shut down
if last_logline_has 1 "(store_errors|faults)" ;then

 cleancache
 echo "Restarting $NUMSQUIDS Frontier Squid${PLURAL}... "
 sleep 3
 $SQUID -DFS
 sleep 3
 isrunning
 RETVAL=$?
 if [ $RETVAL != 0 ] ; then
  echo "Frontier Squid start failed!!!"
  $SQUID -k shutdown 2>/dev/null || true  # in case all are not shut down
 else
  sleep 2
 fi
else
 echo "Script Can't Fix Squid Problem"
 RETVAL=1
fi
 [ $RETVAL == 0 ]
}


stop_squid()
{
 echo -n "Stopping $NUMSQUIDS Frontier Squid${PLURAL}... "
 $SQUID -k shutdown
 RETVAL=$?
 if [ $RETVAL != 0 ] ; then
  echo "Frontier Squid stop failed!!!"
 else
  sleep 30
  echo "done"
 fi
 [ $RETVAL == 0 ]
}

callawstats()
{
 if [ -n "$AWSTATS_SCRIPT" ] && [ -x "$AWSTATS_SCRIPT" ]; then
  for LDIR in $(squid_dirs $ACCESS_LOG_DIR); do
   typeset LFILE=$LDIR/`basename $ACCESS_LOG`
   typeset AWSTATS_CMD="${AWSTATS_SCRIPT} ${LFILE}$1"
   echo "Calling ${AWSTATS_CMD}"
   ${AWSTATS_CMD}
   # break out of this loop because awstats only works on one log file
   # which is OK because multiple squids are only needed on heavily loaded
   #  proxy squid machines where awstats isn't used anyway
   break
  done
 fi
}

rotate()
{
 if [ "$ACCESS_LOG" = none ]; then
  echo "No access log to rotate"
  return
 fi
 echo "Rotating $NUMSQUIDS Frontier Squid Logs... "
 typeset ROTATESUFFIX
 let ROTATESUFFIX="$LOGFILE_ROTATE - 1"
 for D in `squid_dirs $ACCESS_LOG_DIR`; do
  # delete the log files that will be kicked out first, because
  #  otherwise squid stops serving traffic while it removes them
  rm -f $D/access.log.$ROTATESUFFIX $D/cache.log.$ROTATESUFFIX
 done
 $SQUID -k rotate
 RETVAL=$?
 if [ $RETVAL != 0 ]; then
  echo "failed"
 else
  echo "done"
 fi
 callawstats .0
 [ $RETVAL == 0 ]
}

rotateiflarge()
{
 if [ "$ACCESS_LOG" = none ]; then
  rotate
  return
 fi
 # if any access.log is too large, rotate them all
 for LDIR in $(squid_dirs $ACCESS_LOG_DIR); do
  LFILE=$LDIR/`basename $ACCESS_LOG`
  if [ -f "$LFILE" ] && [ "$(stat -c %s $LFILE)" -gt "$LARGE_ACCESS_LOG" ]; then
    rotate
    return
  fi
 done
 echo "It's not necessary to rotate the Frontier Squid Logs"
 callawstats
}

reload()
{
 echo "Reloading $NUMSQUID Frontier Squid configuration${PLURAL}"
 $SQUID -k reconfigure
}

start()
{
 if cleancache; then
  start_squid
 fi
 [ $RETVAL == 0 ]
}

stop()
{
 stop_squid
 [ $RETVAL == 0 ]
}

restart()
{
 if stop; then
  sleep 3
  start_squid   # skip cleaning the cache on a restart
  return $?
 else
  return $?
 fi
}

case "$1" in
  start)
        start
	exit $?
        ;;
  stop)
        stop
	exit $?
        ;;
  status)
        if isrunning; then
	    echo "$NUMSQUIDS Frontier Squid${PLURAL} running"
	    exit 0
	fi 
	echo "Frontier Squid${PLURAL} not running"
        exit 1
        ;;
  restart)
	restart
	exit $?
        ;;
  reload)
	reload
        exit $?
	;;
  condrestart)
	if isrunning; then
	    restart
	    exit $?
	else
	    echo "Frontier Squid not running"
  	fi
        ;;
  cleancache)
	cleancache
	exit $?
	;;
  rotate)
	rotate
	exit $?
	;;
  rotateiflarge)
	rotateiflarge
	exit $?
	;;
  rebuildconf)
	# already done above
        ;;
  *)
        echo $"Usage: $0 {start|stop|restart|condrestart|reload|status|cleancache|rotate|rotateiflarge}"
        exit 1
esac

